@@ -1,4 +1,4 @@
-# $Id: Client-Keepalive.pm 79 2007-02-27 06:07:28Z rcaputo $
+# $Id: Client-Keepalive.pm 104 2008-12-09 06:14:33Z rcaputo $
 
 package POE::Component::Client::Keepalive;
 
@@ -6,10 +6,11 @@ use warnings;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = "0.1000";
+$VERSION = "0.25";
 
 use Carp qw(croak);
 use Errno qw(ETIMEDOUT);
+use Socket qw(SOL_SOCKET SO_LINGER);
 
 use POE;
 use POE::Wheel::SocketFactory;
@@ -23,6 +24,8 @@ eval {
 };
 
 use constant DEBUG => 0;
+use constant DEBUG_DNS => DEBUG || 0;
+use constant DEBUG_DEALLOCATE => DEBUG || 0;
 
 # Manage connection request IDs.
 
@@ -152,21 +155,22 @@ sub new {
     object_states => [
       $self => {
         _start               => "_ka_initialize",
-        _stop                => "_ka_ignore_this_event",
+        _stop                => "_ka_stopped",
+        ka_add_to_queue      => "_ka_add_to_queue",
+        ka_cancel_dns_response => "_ka_cancel_dns_response",
         ka_conn_failure      => "_ka_conn_failure",
         ka_conn_success      => "_ka_conn_success",
+        ka_deallocate        => "_ka_deallocate",
+        ka_dns_response      => "_ka_dns_response",
+        ka_keepalive_timeout => "_ka_keepalive_timeout",
         ka_reclaim_socket    => "_ka_reclaim_socket",
         ka_relinquish_socket => "_ka_relinquish_socket",
         ka_request_timeout   => "_ka_request_timeout",
+        ka_resolve_request   => "_ka_resolve_request",
         ka_set_timeout       => "_ka_set_timeout",
         ka_shutdown          => "_ka_shutdown",
         ka_socket_activity   => "_ka_socket_activity",
-        ka_keepalive_timeout => "_ka_keepalive_timeout",
         ka_wake_up           => "_ka_wake_up",
-        ka_resolve_request   => "_ka_resolve_request",
-        ka_add_to_queue      => "_ka_add_to_queue",
-        ka_dns_response      => "_ka_dns_response",
-        ka_cancel_dns_response => "_ka_cancel_dns_response",
       },
     ],
   );
@@ -183,6 +187,17 @@ sub _ka_initialize {
   $kernel->alias_set("$object");
 }
 
+# When programs crash, the session may stop in a non-shutdown state.
+# _ka_stopped and DESTROY catch this either way the death occurs.
+
+sub _ka_stopped {
+  $_[OBJECT][SF_SHUTDOWN] = 1;
+}
+
+sub DESTROY {
+  my $self = shift;
+  $self->shutdown();
+}
 
 # Request to wake up.  This should only happen during the edge
 # condition where the component's request queue goes from empty to
@@ -238,6 +253,10 @@ sub _ka_wake_up {
         }
       );
 
+      # Remove the wheel-to-request index.
+      delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+      _free_req_id($request->[RQ_ID]);
+
       next;
     }
 
@@ -399,39 +418,61 @@ sub deallocate {
     defined($req_id) and exists($active_req_ids{$req_id})
   );
 
-  my $request = $self->[SF_REQ_INDEX]{$req_id};
+  my $request = delete $self->[SF_REQ_INDEX]{$req_id};
   unless (defined $request) {
-    DEBUG and warn "deallocate could not find request $req_id";
+    DEBUG_DEALLOCATE and warn "deallocate could not find request $req_id";
     return;
   }
+  _free_req_id($request->[RQ_ID]);
+
+  # Now pass the vetted request & its ID into our manager session.
+  $poe_kernel->call("$self", "ka_deallocate", $request, $req_id);
+}
+
+sub _ka_deallocate {
+  my ($self, $heap, $request, $req_id) = @_[OBJECT, HEAP, ARG0, ARG1];
 
   my $conn_key = $request->[RQ_CONN_KEY];
   my $existing_connection = $self->_check_free_pool($conn_key);
 
+  # Existing connection.  Remove it from the pool, and delete the socket.
   if (defined $existing_connection) {
-    # remove it from the pool, delete the socket
     $self->_remove_socket_from_pool($existing_connection->{socket});
-    DEBUG and warn( "deallocate called, deleted already-connected socket" );
-    return;
-  }
-  else {
-    DEBUG and warn(
-      "deallocate called without an existing connection.  ",
-      "cancelling connection request"
+    DEBUG_DEALLOCATE and warn(
+      "deallocate called, deleted already-connected socket"
     );
-
-    my $heap = $poe_kernel->get_active_session->get_heap;
-    $heap->{resolve}->{$request->[RQ_ADDRESS]} = "cancelled";
-
-    $poe_kernel->call( "$self", ka_cancel_dns_response => $request );
     return;
   }
+
+  # No connection yet.  Cancel the request.
+  DEBUG_DEALLOCATE and warn(
+    "deallocate called without an existing connection.  ",
+    "cancelling connection request"
+  );
+
+  unless (exists $heap->{resolve}->{$request->[RQ_ADDRESS]}) {
+    DEBUG_DEALLOCATE and warn(
+      "deallocate cannot cancel dns -- no pending request"
+    );
+    return;
+  }
+
+  if ($heap->{resolve}->{$request->[RQ_ADDRESS]} eq 'cancelled') {
+    DEBUG_DEALLOCATE and warn(
+      "deallocate cannot cancel dns -- request already cancelled"
+    );
+    return;
+  }
+
+  $poe_kernel->call( "$self", ka_cancel_dns_response => $request );
+  return;
 }
 
 sub _ka_cancel_dns_response {
   my ($self, $kernel, $heap, $request) = @_[OBJECT, KERNEL, HEAP, ARG0];
 
   my $address = $request->[RQ_ADDRESS];
+  DEBUG_DNS and warn "DNS: canceling request for $address\n";
   my $requests = $heap->{resolve}{$address};
 
   # Remove the resolver request for the address of this connection
@@ -439,7 +480,6 @@ sub _ka_cancel_dns_response {
 
   my $req_index = @$requests;
   while ($req_index--) {
-    next if $requests->[$req_index] eq "cancelled";
     next unless $requests->[$req_index] == $request;
     splice(@$requests, $req_index, 1);
     last;
@@ -449,7 +489,8 @@ sub _ka_cancel_dns_response {
   # requests to resolve that address.
 
   unless (@$requests) {
-    $heap->{resolve}{$address} = ["cancelled"];
+    DEBUG_DNS and warn "DNS: canceled all requests for $address";
+    $heap->{resolve}{$address} = 'cancelled';
   }
 
   # cancel our attempt to connect
@@ -509,6 +550,7 @@ sub _ka_conn_failure {
 
   # remove the wheel-to-request index
   delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+  _free_req_id($request->[RQ_ID]);
 
   # Discount the use by request key, removing the SF_USED record
   # entirely if it's now moot.
@@ -531,6 +573,7 @@ sub _ka_conn_success {
 
   # remove the wheel-to-request index
   delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+  _free_req_id($request->[RQ_ID]);
 
   # Remove the SF_USED placeholder, add in the socket, and store it
   # properly.
@@ -650,50 +693,61 @@ sub _ka_reclaim_socket {
   my $request_key = $used->[USED_KEY];
   $self->_decrement_used_each($request_key);
 
-  # only try to reuse it if it still works
-  my ($nfound, $status);
-  if (defined fileno($socket)) {
-    DEBUG and warn "RECLAIM: checking if socket still works";
-    my $rin = '';
-    vec($rin, fileno($socket), 1) = 1;
-    my ($rout, $eout);
-    $nfound = select ($rout=$rin, undef, $eout=$rin, 0);
-    DEBUG and warn "RECLAIM: select results: $nfound";
-
-    if ($nfound == -1) {
-      die "select failed: $!";
-    }
-
-    if ($nfound) {
-      use bytes;
-      DEBUG and warn "RECLAIM: uh oh, socket activity";
-      $status = sysread($socket, my $buf = "", 65536);
-      if (DEBUG and defined $status) {
-        warn "RECLAIM: read $status bytes. 0 means EOF";
-      }
-    }
-
-    if (!$nfound or $status != 0) {
-      DEBUG and warn "RECLAIM: reclaiming socket";
-
-      # Watch the socket, and set a keep-alive timeout.
-      $kernel->select_read($socket, "ka_socket_activity");
-      my $timer_id = $kernel->delay_set(
-        ka_keepalive_timeout => $self->[SF_KEEPALIVE], $socket
-      );
-
-      # Record the socket as free to be used.
-      $self->[SF_POOL]{$request_key}{$socket} = $socket;
-      $self->[SF_SOCKETS]{$socket} = [
-        $request_key,       # SK_KEY
-        $timer_id,          # SK_TIMER
-      ];
-    }
-  }
-  else {
+  # Socket is closed.  We can't reuse it.
+  unless (defined fileno $socket) {
     DEBUG and warn "RECLAIM: freed socket has previously been closed";
+    goto &_ka_wake_up;
   }
 
+  # Socket is still open.  Check for lingering data.
+  DEBUG and warn "RECLAIM: checking if socket still works";
+
+  # Check for data on the socket, which implies that the server
+  # doesn't know we're done.  That leads to desynchroniziation on the
+  # protocol level, which strongly implies that we can't reuse the
+  # socket.  In this case, we'll make a quick attempt at fetching all
+  # the data, then close the socket.
+
+  my $rin = '';
+  vec($rin, fileno($socket), 1) = 1;
+  my ($rout, $eout);
+  my $socket_is_active = select ($rout=$rin, undef, $eout=$rin, 0);
+
+  if ($socket_is_active) {
+    DEBUG and warn "RECLAIM: socket is still active; trying to drain";
+    use bytes;
+    my $socket_had_data = sysread($socket, my $buf = "", 65536) || 0;
+    DEBUG and warn "RECLAIM: socket had $socket_had_data bytes. 0 means EOF";
+    DEBUG and warn "RECLAIM: Giving up on socket.";
+
+    # Avoid common FIN_WAIT_2 issues.
+    if ($socket_had_data) {
+      setsockopt($socket, SOL_SOCKET, SO_LINGER, pack("sll",1,0,0)) or die(
+        "setsockopt: $!"
+      );
+    }
+
+    goto &_ka_wake_up;
+  }
+
+  # Socket is alive and has no data, so it's in a quiet, theoretically
+  # reclaimable state.
+
+  DEBUG and warn "RECLAIM: reclaiming socket";
+
+  # Watch the socket, and set a keep-alive timeout.
+  $kernel->select_read($socket, "ka_socket_activity");
+  my $timer_id = $kernel->delay_set(
+    ka_keepalive_timeout => $self->[SF_KEEPALIVE], $socket
+  );
+
+  # Record the socket as free to be used.
+  $self->[SF_POOL]{$request_key}{$socket} = $socket;
+  $self->[SF_SOCKETS]{$socket} = [
+    $request_key,       # SK_KEY
+    $timer_id,          # SK_TIMER
+  ];
+
   goto &_ka_wake_up;
 }
 
@@ -717,6 +771,7 @@ sub _ka_relinquish_socket {
 
 sub shutdown {
   my $self = shift;
+  return if $self->[SF_SHUTDOWN];
   $poe_kernel->call("$self", "ka_shutdown");
 }
 
@@ -741,9 +796,12 @@ sub _ka_shutdown {
 
   # Stop any pending resolver requests.
   foreach my $host (keys %{$heap->{resolve}}) {
+    if ($heap->{resolve}{$host} eq 'cancelled') {
+      DEBUG and warn "SHT: Skipping shutdown for $host (already cancelled)";
+      next;
+    }
     DEBUG and warn "SHT: Shutting down resolver requests for $host";
     foreach my $request (@{$heap->{resolve}{$host}}) {
-      next if $request eq "cancelled";
       $self->_shutdown_request($kernel, $request);
     }
   }
@@ -772,9 +830,10 @@ sub _shutdown_request {
   if (defined $request->[RQ_WHEEL_ID]) {
     DEBUG and warn "SHT: Shutting down resolver wheel $request->[RQ_TIMER_ID]";
     delete $self->[SF_WHEELS]{$request->[RQ_WHEEL_ID]};
-    
+
     # remove the wheel-to-request index
     delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+    _free_req_id($request->[RQ_ID]);
   }
 
   if (defined $request->[RQ_SESSION]) {
@@ -796,10 +855,14 @@ sub _ka_socket_activity {
     warn "CON: Got activity on socket for $key";
   }
 
+  # Any socket activity on a kept-alive socket implies that the socket
+  # is no longer reusable.
+
   use bytes;
-  return if sysread($socket, my $buf = "", 65536);
+  my $socket_had_data = sysread($socket, my $buf = "", 65536) || 0;
+  DEBUG and warn "CON: socket had $socket_had_data bytes. 0 means EOF";
+  DEBUG and warn "CON: Removing socket from the pool";
 
-  DEBUG and warn "CON: socket got EOF or an error. removing it from the pool";
   $self->_remove_socket_from_pool($socket);
 }
 
@@ -807,36 +870,45 @@ sub _ka_resolve_request {
   my ($self, $kernel, $heap, $request) = @_[OBJECT, KERNEL, HEAP, ARG0];
 
   my $host = $request->[RQ_ADDRESS];
-  if ($host !~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
 
-    if (exists $heap->{resolve}->{$host}) {
-      DEBUG and warn "DNS: $host is piggybacking on a pending lookup.\n";
-      push @{$heap->{resolve}->{$host}}, $request;
-    } else {
-      DEBUG and warn "DNS: $host is being looked up in the background.\n";
-      $heap->{resolve}->{$host} = [ $request ];
-
-      my $response = $self->[SF_RESOLVER]->resolve(
-        event   => 'ka_dns_response',
-        host    => $host,
-        context => 1,
-      );
-
-      if ($response) {
-        DEBUG and warn "DNS: resolver returned immediately";
-        $kernel->yield (ka_dns_response => $response);
-      }
-    }
-  } else {
-    DEBUG and warn "DNS: $host may block while it's looked up.\n";
+  # Skip DNS resolution if it's already a dotted quad.
+  # TODO - Not all dotted quads are good.
+  if ($host =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+    DEBUG_DNS and warn "DNS: $host is a dotted quad; skipping lookup";
     $kernel->call("$self", ka_add_to_queue => $request);
+    return;
   }
+
+  # It's already pending DNS resolution.  Combine this with previous.
+  if (exists $heap->{resolve}->{$host}) {
+    DEBUG_DNS and warn "DNS: $host is piggybacking on a pending lookup.\n";
+    push @{$heap->{resolve}->{$host}}, $request;
+    return;
+  }
+
+  # New request.  Start lookup.
+  $heap->{resolve}->{$host} = [ $request ];
+
+  my $response = $self->[SF_RESOLVER]->resolve(
+    event   => 'ka_dns_response',
+    host    => $host,
+    context => 1, # required but unused
+  );
+
+  if ($response) {
+    DEBUG_DNS and warn "DNS: immediate resolution for $host";
+    $kernel->yield(ka_dns_response => $response);
+    return;
+  }
+
+  DEBUG_DNS and warn "DNS: looking up $host in the background.\n";
 }
 
 sub _ka_dns_response {
   my ($self, $kernel, $heap, $response) = @_[OBJECT, KERNEL, HEAP, ARG0];
 
-  # Nothing to do if we're shut down.
+  # We've shut down.  Nothing to do here.
+  return if $self->[SF_SHUTDOWN];
 
   my $request_address = $response->{'host'};
   my $response_object = $response->{'response'};
@@ -844,44 +916,43 @@ sub _ka_dns_response {
 
   my $requests = delete $heap->{resolve}->{$request_address};
 
-  DEBUG and warn "DNS: request address = $request_address";
+  DEBUG_DNS and warn "DNS: got response for request address $request_address";
 
-  # No requests are on record for this lookup.
-  # TODO - What if the requests were canceled, possibly due to a
-  # timeout?  The answer: Timeouts don't cancel DNS requests, so we
-  # still get the response (eventually) and clean up the resolver
-  # thingy here.  DNS request piggybacking should prevent this from
-  # being called multiple times, so we should be safe there, too.
-  # Leave the die() in until coverage == 100%, so we can catch
-  # possible bugs in the future.
-  #
-  # If it says "cancelled", it was.
-  return unless defined $requests;
+  # Requests on record.
   if (defined $requests) {
-    if (ref $requests eq 'ARRAY') {
-        return if grep { $_ eq 'cancelled' } @$requests;
-    } else {
-      die "!!!: got a defined requests but NOT an arrayref";
+    # We can receive responses for canceled requests.  Ignore them: we
+    # cannot cancel PoCo::Client::DNS requests, so this is how we reap
+    # them when they're canceled.
+    if ($requests eq 'cancelled') {
+      DEBUG_DNS and warn "DNS: reaping cancelled request for $request_address";
+      return;
     }
-  } else {
-     die "!!!: Unexpectedly undefined requests";
+    unless (ref $requests eq 'ARRAY') {
+      die "DNS: got an unknown requests for $request_address: $requests";
+    }
+  }
+  else {
+    die "DNS: Unexpectedly undefined requests for $request_address";
   }
 
   # No response.  This is an error.  Cancel all requests for the
   # address.  Tell everybody that their requests timed out.
   unless (defined $response_object) {
+    DEBUG_DNS and warn "DNS: undefined response = error";
     foreach my $request (@$requests) {
       _respond_with_error($request, "resolve", undef, $response_error),
     }
     return;
   }
 
+  DEBUG_DNS and warn "DNS: got a response";
+
   # A response!
   foreach my $answer ($response_object->answer()) {
     # don't need this because we ask for only A answers anyway
     #next unless $answer->type eq "A";
 
-    DEBUG and warn "DNS: $request_address resolves to ", $answer->rdatastr;
+    DEBUG_DNS and warn "DNS: $request_address resolves to ", $answer->rdatastr;
 
     foreach my $request (@$requests) {
       # Don't bother continuing inactive requests.
@@ -896,7 +967,7 @@ sub _ka_dns_response {
 
   # Didn't return here.  No address record for the host?
   foreach my $request (@$requests) {
-    DEBUG and warn "DNS: $request_address does not resolve";
+    DEBUG_DNS and warn "DNS: $request_address does not resolve";
     _respond_with_error($request, "resolve", undef, "Host has no address."),
   }
 }
@@ -926,7 +997,7 @@ sub _ka_add_to_queue {
   return if (
     ($self->[SF_USED_EACH]{$conn_key} || 0) >= $self->[SF_MAX_HOST]
   );
-  
+
   # Wake the session up, and return nothing, signifying sound and fury
   # yet to come.
   DEBUG and warn "posting wakeup for $conn_key";
@@ -952,6 +1023,19 @@ sub _remove_socket_from_pool {
 
   $poe_kernel->alarm_remove($socket_rec->[SK_TIMER]);
   $poe_kernel->select_read($socket, undef);
+
+  # Avoid common FIN_WAIT_2 issues.
+  # Commented out because fileno() will return true for closed
+  # sockets, which makes setsockopt() highly unhappy.  Also, SO_LINGER
+  # will cause te socket closure to block, which is less than ideal.
+  # We need to revisit this another way, or just let sockets enter
+  # FIN_WAIT_2.
+
+#  if (fileno $socket) {
+#    setsockopt($socket, SOL_SOCKET, SO_LINGER, pack("sll",1,0,0)) or die(
+#      "setsockopt: $!"
+#    );
+#  }
 }
 
 # Internal function.  NOT AN EVENT HANDLER.
@@ -973,7 +1057,7 @@ sub _respond {
   my ($request, $fields) = @_;
 
   # Bail out early if the request isn't active.
-  return unless $request->[RQ_ACTIVE];
+  return unless $request->[RQ_ACTIVE] and $request->[RQ_SESSION];
 
   $poe_kernel->post(
     $request->[RQ_SESSION],
